# -*- coding: utf-8 -*-
#
# This file is part of the imio.scan_helpers distribution (https://github.com/IMIO/imio.scan_helpers).
# Copyright (c) 2023 IMIO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from config import CERT_FILE
from config import COPY_BAT_NAME
from config import DOWNLOAD_DIR
from config import get_bundle_dir
from config import GITHUB_REPO
from config import IS_PROD
from config import MAIN_BACKUP_DIR
from config import MAIN_EXE_NAME
from config import PROFILES_DIRS
from logger import close_logger
from logger import log

import json
import os
import re
import requests
import shutil
import subprocess
import sys
import traceback
import zipfile


def copy_sub_files(src_dir, dest_dir, files=()):
    """Copy content from src to dest"""
    for item in os.listdir(src_dir):
        if files and item not in files:
            continue
        s = os.path.join(src_dir, item)
        d = os.path.join(dest_dir, item)
        log.info(f'Copying "{s}" to "{dest_dir}"')
        if os.path.isdir(s):
            if os.path.exists(d):
                shutil.rmtree(d)
            shutil.copytree(s, d)
        else:
            shutil.copy2(s, d)


def copy_release_files_and_restart(src_dir, dest_dir):
    """Will create a bat to copy files after main process has ended and restart the main process without upgrade"""
    exe_path = os.path.join(dest_dir, f"{MAIN_EXE_NAME}.exe")
    script_path = os.path.join(dest_dir, COPY_BAT_NAME)

    with open(script_path, "w") as script:
        script.write("@echo off\n")
        script.write(f'echo Copying "{src_dir}" files to "{dest_dir}""\n')
        script.write("timeout /t 3\n")  # waits for main script to end
        script.write(f'xcopy /s /e /h /r /y /q "{src_dir}\\*" "{dest_dir}"\n')
        script.write(f'start "" "{exe_path}" -nu\n')
        script.write(f'rmdir /s /q "{src_dir}"\n')
        # script.write(f'del "{script_path}"\n')

    if IS_PROD:
        subprocess.Popen(["cmd", "/c", script_path])


def download_update(url, download_path):
    """Download github zip file"""
    response = requests.get(url, stream=True)
    response.raise_for_status()
    with open(download_path, "wb") as file:
        for chunk in response.iter_content(chunk_size=8192):
            file.write(chunk)


def exception_infos(ex):
    """Get exception infos"""
    tb_list = traceback.extract_tb(ex.__traceback__)
    filename, line_number, func_name, text = tb_list[-1]
    return f"in {filename} at line {line_number}: '{ex}'"


def get_dated_backup_dir(backup_dir, day):
    """Get dated backup dir"""
    dated_dir = os.path.join(backup_dir, day)
    if not os.path.exists(dated_dir):
        os.makedirs(dated_dir)
    return dated_dir


def get_download_dir_path():
    """Get download dir path for new version zip files"""
    # maybe use tempdir
    # temp_dir = tempfile.gettempdir()
    return os.path.join(get_bundle_dir(), DOWNLOAD_DIR)


def get_last_dated_backup_dir(backup_dir):
    """Get last dated backup dir"""
    subdirs = read_dir(backup_dir, with_path=False, only_folders=True)
    subdirs = [adir for adir in subdirs if re.match(r"^\d{4}-\d{2}-\d{2}$", adir)]
    if subdirs:
        subdirs.sort(reverse=True)
        return os.path.join(backup_dir, subdirs[0])
    return None


def get_latest_release_version(params, release=None):
    """Get GitHub latest or specified release info"""
    if release:
        url = f"https://api.github.com/repos/{GITHUB_REPO}/releases"
        ret = json_request(url, params)
        if not ret:
            return None, None
        for dic in ret:
            if dic["tag_name"] == release:
                url = f"https://api.github.com/repos/{GITHUB_REPO}/releases/{dic['id']}"
                break
        else:
            stop(f"The release with tag '{release}' cannot be found", params=params)
    else:
        url = f"https://api.github.com/repos/{GITHUB_REPO}/releases/latest"
    latest_release = json_request(url, params)
    if not latest_release:
        return None, None
    return latest_release["tag_name"], latest_release["assets"][0]["browser_download_url"]


def get_main_backup_dir(create=True):
    """Get main backup dir"""
    if not os.path.exists(MAIN_BACKUP_DIR):
        if create:
            os.makedirs(MAIN_BACKUP_DIR)
        else:
            return None
    return MAIN_BACKUP_DIR


def get_parameter(params_file, param=None, default=None):
    """Get a specific parameter or the full dic"""
    dic = {}
    if os.path.exists(params_file):
        with open(params_file) as pf:
            dic = json.load(pf)
    if param:
        return dic.get(param, default)
    else:
        return dic


def get_scan_profiles_dir():
    """Get profile dir to read or write"""
    for prof_dir in PROFILES_DIRS:
        if os.path.exists(prof_dir):
            return prof_dir


def json_request(url, params):
    """Simple json request"""
    proxies = {}
    if params.get("PROXY"):
        for protocol in ["http", "https"]:
            if params.get("PROXY_USER") and params.get("PROXY_PWD"):
                proxies[protocol] = f"{params['PROXY_USER']}:{params['PROXY_PWD']}@{params['PROXY']}"
            else:
                proxies[protocol] = params["PROXY"]
    try:
        response = requests.get(url, proxies=proxies, verify=CERT_FILE)
        response.raise_for_status()
    except requests.exceptions.ProxyError as err:
        send_log_message(f"Cannot request '{url}' : '{err}' with proxies {proxies}", params)
        return {}
    except Exception as err:
        send_log_message(f"Cannot request '{url}' : '{err}'", params)
        return {}
    return response.json()


def read_dir(dirpath, with_path=False, only_folders=False, only_files=False, to_skip=[]):
    """Read the dir and return files"""
    files = []
    for filename in os.listdir(dirpath):
        if filename in to_skip:
            continue
        if only_folders and not os.path.isdir(os.path.join(dirpath, filename)):
            continue
        if only_files and not os.path.isfile(os.path.join(dirpath, filename)):
            continue
        if with_path:
            files.append(os.path.join(dirpath, filename))
        else:
            files.append(filename)
    return files


def send_log_message(message, params, log_method=log.error, level="ERROR"):
    """Send log message to webservice.

    :param message: The message to send
    :param params: The parameters dic containing webservice info
    :param log_method: The log method to use
    :param level: The log level to include: ERROR or INFO
    """
    data = {
        "client_id": params["CLIENT_ID"],
        "hostname": params["hostname"],
        "message": message,
        "level": level,
        "version": params.get("version"),
    }
    if log_method:
        log_method(message)
    headers = {"Content-Type": "application/json", "Accept": "application/json"}
    response = requests.post(
        params["SERVER_URL"], headers=headers, data=json.dumps(data), auth=("loguser", params["PLONE_PWD"])
    )
    if response.status_code != 200:
        log.error(f"Failed to send log message: {response.text}")


def set_parameter(params_file, key, value):
    """Store client_id in file"""
    if key == "CLIENT_ID" and not re.match(r"^0\d{5}$", value):
        stop(f"Given client_id '{value}' not well formed !")
    dic = get_parameter(params_file)
    dic[key] = value
    with open(params_file, "w") as pf:
        json.dump(dic, pf)
    return dic


def stop(msg="", intup=True, params=None, exit_code=1):
    """Stop the script and log the message.

    :param msg: The message to log
    :param intup: Flag to wait for input before closing
    :param params: The parameters dic containing webservice info. If given the message is sent to the webservice
    :param exit_code: The exit code to return
    """
    if msg:
        log.error(msg)
        if params:
            send_log_message(msg, params, log_method=None)
    if intup:
        input("Press Enter to exit...")
    close_logger()
    sys.exit(exit_code)


def unzip_file(zip_path, extract_to):
    """Unzip downloaded archive and delete it"""
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(extract_to)
    os.remove(zip_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of the imio.scan_helpers distribution (https://github.com/IMIO/imio.scan_helpers).
# Copyright (c) 2023 IMIO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from config import get_bundle_dir
from config import get_current_version
from config import MAIN_EXE_NAME
from config import PARAMS_FILE_NAME
from config import SERVER_URL
from logger import close_logger
from logger import log
from profiles_restore import main as profiles_restore_main
from utils import copy_release_files_and_restart
from utils import download_update
from utils import exception_infos
from utils import get_download_dir_path
from utils import get_last_dated_backup_dir
from utils import get_latest_release_version
from utils import get_main_backup_dir
from utils import get_parameter
from utils import read_dir
from utils import send_log_message
from utils import set_parameter
from utils import stop
from utils import unzip_file

import argparse
import os
import platform


def handle_startup(main_dir, params, action="add"):
    """Add/remove exe to/from startup"""
    exe_path = os.path.join(main_dir, f"{MAIN_EXE_NAME}.exe")
    key = r"Software\Microsoft\Windows\CurrentVersion\Run"
    value_name = "IMIO_Scan_Helpers_Scripts"
    try:
        import winreg

        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, key, 0, winreg.KEY_SET_VALUE) as reg_key:
            if action == "add":
                winreg.SetValueEx(reg_key, value_name, 0, winreg.REG_SZ, f'"{exe_path}" --is-auto-started')
                log.info(f"'{exe_path} --is-auto-started' added to startup")
            elif action == "remove":
                winreg.DeleteValue(reg_key, value_name)
                log.info(f"'{exe_path}' removed from startup")
    except ImportError:
        send_log_message("Cannot import winreg: add to startup failed !!", params)
    except Exception as e:
        send_log_message(f"Error in handle_startup, {exception_infos(e)}", params)


def check_for_updates(main_dir, cur_vers, params):
    """Check for updates"""
    latest_version, download_url = get_latest_release_version(params, ns.release)
    if latest_version is not None and (latest_version > cur_vers or ns.release):
        log.info(f"New version available: {latest_version}")
        download_dir_path = get_download_dir_path()
        if not os.path.exists(download_dir_path):
            os.makedirs(download_dir_path)
        download_path = os.path.join(download_dir_path, download_url.split("/")[-1])
        log.info(f"Downloading {download_url} to {download_path}")
        download_update(download_url, download_path)
        log.info(f"Unzipping {download_path} to {download_dir_path}")
        unzip_file(download_path, download_dir_path)
        copy_release_files_and_restart(download_dir_path, main_dir)
        log.info("Will replace files and restart")
        stop(intup=False, exit_code=0)


# Argument parsing
parser = argparse.ArgumentParser()
parser.add_argument("-v", "--version", action="store_true", dest="version", help="Show version")
parser.add_argument("-c", "--client-id", dest="client_id", help="Set client id")
parser.add_argument("-p", "--password", dest="password", help="Set password")
parser.add_argument("-nu", "--no-update", action="store_true", dest="no_update", help="Do not check for updates")
parser.add_argument("-r", "--release", dest="release", help="Get this release")
parser.add_argument("-tm", "--test-message", action="store_true", dest="test_message", help="Send a test message")
parser.add_argument("--proxy", dest="proxy", help="Proxy url")
parser.add_argument("--proxy-user", dest="proxy_user", help="Proxy user")
parser.add_argument("--proxy-pwd", dest="proxy_pwd", help="Proxy password")
parser.add_argument("--startup", action="store_true", dest="startup", help="Add exe to startup")
parser.add_argument("--startup-remove", action="store_true", dest="startup_remove", help="Remove exe from startup")
parser.add_argument(
    "--is-auto-started", action="store_true", dest="auto_started", help="Flag when script is auto started"
)
ns = parser.parse_args()

if ns.version:
    print(f"imio.scan_helpers version {get_current_version()}")
    stop(intup=False, exit_code=0)
bundle_dir = get_bundle_dir()
log.info(f"dir={bundle_dir}")

params_file = os.path.join(bundle_dir, PARAMS_FILE_NAME)
if ns.client_id:
    set_parameter(params_file, "CLIENT_ID", ns.client_id)
if ns.password:
    set_parameter(params_file, "PLONE_PWD", ns.password)
if ns.proxy:
    set_parameter(params_file, "PROXY", ns.proxy)
if ns.proxy_user:
    set_parameter(params_file, "PROXY_USER", ns.proxy_user)
if ns.proxy_pwd:
    set_parameter(params_file, "PROXY_PWD", ns.proxy_pwd)
parameters = set_parameter(params_file, "hostname", platform.node())
if "CLIENT_ID" not in parameters or "PLONE_PWD" not in parameters:
    stop("CLIENT_ID or PLONE_PWD not found in parameters")
if "SERVER_URL" not in parameters:
    parameters = set_parameter(params_file, "SERVER_URL", SERVER_URL)
current_version = get_current_version()
if "version" not in parameters:
    parameters = set_parameter(params_file, "version", current_version)
if current_version != parameters["version"]:
    old_version = parameters["version"]
    parameters = set_parameter(params_file, "version", current_version)
    send_log_message(
        f"Product updated from {old_version} to {current_version}",
        parameters,
        log_method=log.info,
        level="INFO",
    )

try:
    if ns.startup:
        handle_startup(bundle_dir, parameters)
    if ns.startup_remove:
        handle_startup(bundle_dir, parameters, action="remove")
    if ns.no_update or ns.test_message:
        pass
    else:
        check_for_updates(bundle_dir, current_version, parameters)
except Exception as ex:
    send_log_message(f"General error in main script, {exception_infos(ex)}", parameters)

# will do something
log.info(f"Current version is {get_current_version()}")
main_backup_dir = get_main_backup_dir(create=False)
if ns.auto_started and not get_parameter(params_file, "various"):
    back_dirs = []
    if main_backup_dir:
        dated_backup_dir = get_last_dated_backup_dir(main_backup_dir)
        if dated_backup_dir:
            back_dirs = read_dir(dated_backup_dir, with_path=False, only_folders=True)
    send_log_message(
        f"Script started automatically !\nInstalled in '{bundle_dir}'.\n"
        f"Version is {parameters['version']}\nParameters: {sorted(parameters.keys())}\n"
        f"Backuped dirs: {back_dirs}",
        parameters,
        log_method=log.info,
    )
    set_parameter(params_file, "various", "auto_started")
if ns.test_message:
    send_log_message(
        f"Test message\nInstalled in '{bundle_dir}'.",
        parameters,
        log_method=log.info,
    )
elif get_main_backup_dir(create=False):
    profiles_restore_main()
close_logger()
